#ifndef VECTORDUMB_H
#define VECTORDUMB_H


#include <iostream.h>
#include "Alloc.h"

template <class T, int N>
class Vector {
 private:
    static Allocator<T, N> allocator_data;

    T* dataStart;
    T* dataFinish;

 public:
    Inline T* begin(void) const ;
    Inline T* end(void) const;
    Inline int size(void) const;
    Inline T& operator[](int n);
    Inline Vector();
    Inline Vector(const Vector<T,N>& x);
    Inline ~Vector();
    Inline Vector<T,N>& operator=(const Vector<T,N>& that);
    Inline Vector<T,N> operator+(const Vector<T,N>& v);
    Inline Vector<T,N>& operator+=(const Vector<T,N>& v);
    Inline Vector<T,N> operator-(const Vector<T,N>& v);
    Inline Vector<T,N>& operator-=(const Vector<T,N>& v);
    Inline T operator*(const Vector<T,N>& v);
    Inline Vector<T,N> operator*(T t);
};

template <class T, int N>
Inline T* Vector<T,N>::begin(void) const  
{ 
  return dataStart; 
}

template <class T, int N>
Inline T* Vector<T,N>::end(void) const  
{ 
  return dataFinish; 
}

template <class T, int N>
Inline int Vector<T,N>::size(void) const 
{ 
  return N;
}

template <class T, int N>
Inline T& Vector<T,N>::operator[](int n) 
{ 
  return dataStart[n]; 
}


template <class T, int N>
Inline Vector<T,N>::Vector() 
: dataStart(allocator_data.allocate()), 
  dataFinish(dataStart+N)
{
}

template <class T, int N>
Inline Vector<T,N>::~Vector() 
{ 
  allocator_data.deallocate(dataStart);
}


template <class T, int N>
Inline Vector<T,N> Vector<T,N>::operator+(const Vector<T,N>& v)
{
    Vector<T,N> temp(*this);
    temp+=v;
    return temp;
}



template <class T, int N>
Inline Vector<T,N> Vector<T,N>::operator-(const Vector<T,N>& v)
{
    Vector<T,N> temp(*this);
    temp-=v;
    return temp;
}

 
#ifdef NOSPECIAL 
template <class T, int N>
Inline Vector<T,N>::Vector(const Vector<T,N>& x) 
: dataStart(allocator_data.allocate()), 
  dataFinish(dataStart+N)
{
  T* dest = dataStart;
  T* src = x.dataStart;
  while (src!=x.dataFinish)
    *dest++=*src++;
}

template <class T, int N>
Inline Vector<T,N>& Vector<T,N>::operator=(const Vector<T,N>& that)
{
  if (&that == this) return *this;
  T* dest = dataStart;
  T* src = that.dataStart;
  while (src!=that.dataFinish)
    *dest++=*src++;
  return *this;
}

template <class T, int N>
Inline Vector<T,N>& Vector<T,N>::operator+=(const Vector<T,N>& v)
{
    T* dest=dataStart;
    T* src = v.dataStart;
    while (dest!=dataFinish)
        *(dest++)+=*(src++);
    return *this;
}

template <class T, int N>
Inline Vector<T,N>& Vector<T,N>::operator-=(const Vector<T,N>& v)
{
    T* dest=dataStart;
    T* src = v.dataStart;
    while (dest!=dataFinish)
        *(dest++)-=*(src++);
    return *this;
}

template <class T, int N>
Inline T Vector<T,N>::operator*(const Vector<T,N>& v)
{
    double result=0;
    T* dest=dataStart;
    T* src = v.dataStart;
    while (dest!=dataFinish)
        result+=(*dest++)*(*src++);
    return result;
}

template <class T, int N>
Inline Vector<T, N> Vector<T, N>::operator*(T t)
{
  Vector<T, N> tv(*this);

  T* ptr(tv.begin());
  do
    *(ptr++) *= t;
  while (ptr != tv.end());
  
  return tv;
}

#else

Inline Vector<double,3>::Vector(const Vector<double,3>& x) 
: dataStart(allocator_data.allocate()), 
  dataFinish(dataStart+3)
{
  double* dest = dataStart;
  double* src = x.dataStart;
  *dest++=*src++;
  *dest++=*src++;
  *dest=*src;
}

Inline Vector<double,3>& Vector<double,3>::operator=(const Vector<double,3>& that)
{
  if (&that == this) return *this;
  double* dest = dataStart;
  double* src = that.dataStart;
  *dest++=*src++;
  *dest++=*src++;
  *dest=*src;
  return *this;
}

Inline Vector<double,3>& Vector<double,3>::operator+=(const Vector<double,3>& v)
{
    double* dest=dataStart;
    double* src = v.dataStart;
    *(dest++)+=*(src++);
    *(dest++)+=*(src++);
    *(dest)+=*(src);
    return *this;
}

Inline Vector<double,3>& Vector<double,3>::operator-=(const Vector<double,3>& v)
{
    double* dest=dataStart;
    double* src = v.dataStart;
    *(dest++)-=*(src++);
    *(dest++)-=*(src++);
    *(dest)-=*(src);
    return *this;
}

Inline double Vector<double,3>::operator*(const Vector<double,3>& v)
{
    double result=0;
    double* dest=dataStart;
    double* src = v.dataStart;
    result+=(*dest++)*(*src++);
    result+=(*dest++)*(*src++);
    result+=(*dest)*(*src);
    return result;
}


Inline Vector<double, 3> Vector<double, 3>::operator*(double t)
{
  Vector<double,3> tv(*this);

  double* ptr(tv.begin());
  do
    *(ptr++) *= t;
  while (ptr != tv.end());
  
  return tv;
}

#endif

#endif

  
// Local Variables:
// mode: c++
// End:



