#ifndef ALLOCS_H
#define ALLOCS_H


template <class T, int N>
class Allocator
{
private:
  union {
    T content[N];
    Allocator<T,N>* next;
    } magic;

public:

  Inline Allocator();
  Inline Allocator(int n);
  Inline T* allocate(void);
  Inline void deallocate(const T* s);
  Inline ~Allocator();
};

template <class T, int N>
Inline  Allocator<T,N>::Allocator()
{
  magic.next=NULL;
}

template <class T, int N>
Inline  Allocator<T,N>::Allocator(int n)
{
  Allocator<T,N>* block = new Allocator<T,N>[n];
  for (int i=0;i<n-1;++i)
    block[i].magic.next=&(block[i+1]);
  block[n-1].magic.next=NULL;
  magic.next=block;
}


template <class T, int N>
Inline T*  Allocator<T,N>::allocate(void)
{
  if (magic.next)
    {
      T* s=magic.next->magic.content;
      magic.next=magic.next->magic.next;
      return s;
    }
  else
    {
      Allocator<T,N>* tmp = new Allocator<T,N>;
      return tmp->magic.content;
    }
}


template <class T, int N>
Inline void Allocator<T,N>::deallocate(const T* s)
{
  Allocator<T,N>* tmp = (Allocator<T,N>*) s;
  tmp->magic.next=magic.next;
  magic.next=tmp;
}

template <class T, int N>
Inline  Allocator<T,N>::~Allocator()
{
  if (magic.next)
    {
      delete magic.next; // recursive !
    }
}

#endif 

// Local Variables:
// mode: c++
// End:


