#include <stdlib.h>
#include <math.h>
#include <iostream.h>
#include <stdlib.h>

#include "Vector.h"

Allocator<double, 3> TVector<double,3>::allocator_data;

double ZERO=0.0;
double ONE=1.0;


#ifdef NAG
extern "C" double g05daf(double* a, double* b);
#define drand48() g05daf(&ZERO,&ONE)
#endif


class Sigma
{
public:
    int T, L, N;
    //
    // index fields for neighbors
    // 

    Vector<double,3> *sigma;
    Vector<double,3> sigmaSum,sigmaTrial;

    double beta,deltaS;

    int *neighborN, *neighborW, *neighborS, *neighborE;

    int accept, reject;

    Sigma(int timeExtension,
	  int spaceExtension,
	  double b);
    ~Sigma();

    void randomize3(Vector<double,3>& s);

    int index(int t, int x); 
    void sumNeighbors(int i);
    void metropolisUpdate(int sweeps, int hits);
    void localMetropolisUpdate(int i, int hits);
    double measureLinkEnergy();
};


inline int Sigma::index(int t, int x)
{
    return L*t+x;
}

void Sigma::randomize3(Vector<double,3>& s)
{
    double r;
    do { 
        s[0]=2.0*drand48()-1.0; 
        s[1]=2.0*drand48()-1.0; 
        s[2]=2.0*drand48()-1.0;
        r=s[0]*s[0]+s[1]*s[1]+s[2]*s[2]; 
    } while (r >= 1.0 || r == 0.0);
    r=sqrt(r);
    s[0] /= r;
    s[1] /= r;
    s[2] /= r;
}

Sigma::Sigma(int timeExtension,
	     int spaceExtension,
	     double b)
:       T(timeExtension),
	L(spaceExtension),
      N(L*T),
      beta(b),
      accept(0),
      reject(0)
{

  sigma = new Vector<double,3>[N];
  sigmaSum =Vector<double,3>(); 
  sigmaTrial = Vector<double,3>();
    
    neighborE = new int[N];
    neighborN = new int[N];
    neighborW = new int[N];
    neighborS = new int[N];

    for (int t=0;t<T;t++)
	for (int x=0;x<L;x++)
	    {
		int i = index(t,x);
		sigma[i]=Vector<double,3>();

		randomize3(sigma[i]);
			   
	        int tup = (t+1)%T;
		int tdown = (t-1+T)%T;
		int xup = (x+1)%L;
		int xdown = (x-1+L)%L;


		
		neighborE[i]=index(tup,x);
		neighborN[i]=index(t,xup);
		neighborW[i]=index(tdown,x);
		neighborS[i]=index(t,xdown);
	    }
}


Sigma::~Sigma()
{
     delete [] sigma;

     delete [] neighborE;
     delete [] neighborN;
     delete [] neighborW;
     delete [] neighborS;
}

void Sigma::sumNeighbors(int i)
{
    int iE = neighborE[i];		
    int iN = neighborN[i];		
    int iW = neighborW[i];		
    int iS = neighborS[i];		
    
    sigmaSum=sigma[iE]+sigma[iN]+sigma[iW]+sigma[iS];
}

void Sigma::localMetropolisUpdate(int i, int hits)
{
    sumNeighbors(i);

    while (hits--)
	{
	    randomize3(sigmaTrial);
	    deltaS = 
		-beta * ((sigmaTrial-sigma[i])*sigmaSum);

	    if (deltaS<0 || 
		drand48() < exp(-deltaS))
		{
		    sigma[i]=sigmaTrial;
		    accept++;
		}
	    else
		{
		    reject++;
		}

	}
}


void Sigma::metropolisUpdate(int sweeps, int hits)
{
  while (sweeps--)
    {
      for (int i=0;i<N;i++)
	  localMetropolisUpdate(i,hits);
    }
}

double Sigma::measureLinkEnergy()
{
    double link=0;
    for (int t=0;t<T;t++)
	for (int x=0;x<L;x++)
	    {
		int i = index(t,x);
		int iE = neighborE[i];		
		int iN = neighborN[i];		
		int iW = neighborW[i];		
		int iS = neighborS[i];		

		link+=
		    (sigma[iE]+sigma[iN]+sigma[iW]+sigma[iS])*sigma[i];
	    }
    link /= double(N);
    return link;
}


int main()
{
    int equi=1000;
    int sweeps=1000;
    Sigma sigma(10,10,1.0);
    sigma.metropolisUpdate(equi,4);
    double meanLink=0;
    for (int s=0;s<sweeps;s++)
	{
	    sigma.metropolisUpdate(1,4);
	    meanLink+=sigma.measureLinkEnergy();
	}
    meanLink /= double(sweeps);
    cout << "mean=" << meanLink << endl;
    cout << "ratio=" << double(sigma.accept)/double(sigma.accept+sigma.reject) << endl;

    return 0;
  }



