#ifndef VECTORTEMP_H
#define VECTORTEMP_H

#include "Alloc.h"

template <class T, int N> class Vector;

template <class T, int N>
class TVector
{
  friend class Vector<T,N>;
    
 protected:

  T* dataStart;
  T* dataFinish;

  static Allocator<T, N> allocator_data;
public:

   TVector() : dataStart(allocator_data.allocate()),
                     dataFinish(dataStart+N) { };
   TVector(const TVector<T, N>& tv) { dataStart=tv.dataStart;
					    dataFinish=tv.dataFinish;
					  }
  TVector(T* start);
   ~TVector() {};
  
   TVector<T, N> operator+(const Vector<T, N>& v) const;
   TVector<T, N> operator+(const TVector<T, N>& tv) const;

   TVector<T, N> operator-(const Vector<T, N>& v) const;
   TVector<T, N> operator-(const TVector<T, N>& tv) const;

   TVector<T, N> operator*(T t) const;

   T              operator*(const Vector<T, N>& v) const;
   T              operator*(const TVector<T, N>& tv) const;  
};
 


template <class T, int N>
class Vector : public TVector<T, N>
{
 public:
   Vector() {};
   Vector(const Vector<T, N>& tv);
   ~Vector(){ allocator_data.deallocate(dataStart);}
  T&            operator[](int i) {return dataStart[i];}

   Vector<T, N>& operator=(const Vector<T, N>& v);
   Vector<T, N>& operator=(const TVector<T, N>& tv);

   TVector<T, N> operator+(const Vector<T, N>& v) const;
         TVector<T, N> operator+(const TVector<T, N>& tv) const
	    { return tv+(*this); }

   TVector<T, N> operator-(const Vector<T, N>& v) const;
         TVector<T, N> operator-(const TVector<T, N>& tv) const
	   { return tv-(*this); }

   TVector<T, N> operator*(T t) const;
   T             operator*(const Vector<T, N>& v) const;
   T             operator*(const TVector<T, N>& tv) const;
};



#ifndef NOSPECIAL

Inline TVector<double,3>
TVector<double, 3>::operator+(const TVector<double, 3>& tv) const
{
  double* sum1(dataStart);
  double* sum2(tv.dataStart);

  *(sum1++) += *(sum2++); 
  *(sum1++) += *(sum2++);     
  *sum1     += *sum2;
  
  allocator_data.deallocate(tv.dataStart);
  
  return *this;
}



Inline TVector<double,3>
TVector<double, 3>::operator+(const Vector<double, 3>& v) const
{
  
  double* sum1(dataStart);
  double* sum2(v.dataStart);

  *(sum1++) += *(sum2++);    
  *(sum1++) += *(sum2++);
  *sum1     += *sum2;

  return *this;
}


Inline TVector<double,3>
TVector<double, 3>::operator-(const TVector<double, 3>& tv) const
{
  double* sum1(dataStart);
  double* sum2(tv.dataStart);

  *(sum1++) -= *(sum2++); 
  *(sum1++) -= *(sum2++);     
  *sum1     -= *sum2;
  
  allocator_data.deallocate(tv.dataStart);
  
  return *this;
}

Inline TVector<double, 3> TVector<double,3>::operator*(double t) const
{
  double* ptr(dataStart);

  *(ptr++) *= t;
  *(ptr++) *= t;
  *(ptr) *= t;
  
  return *this;
}


Inline Vector<double, 3>::Vector(const Vector<double, 3>& v)
           : TVector<double, 3>(allocator_data.allocate())
{
  double* dest(dataStart);
  double* src(v.dataStart);

  *(dest++)=*(src++);
  *(dest++)=*(src++);
  *dest = *src;
}



Inline Vector<double, 3>&
Vector<double, 3>::operator=(const Vector<double, 3>& v)
{
  double* dest(dataStart);
  double* src(v.dataStart);

  *(dest++)=*(src++);
  *(dest++)=*(src++);
  *dest = *src;

  return *this;
}


Inline Vector<double, 3>&
Vector<double, 3>::operator=(const TVector<double, 3>& tv)
{
  allocator_data.deallocate(dataStart);

  dataStart=tv.dataStart;
  dataFinish  =tv.dataFinish;

  return *this;
}


Inline TVector<double, 3>
Vector<double, 3>::operator+(const Vector<double, 3>& v) const
{
  TVector<double, 3> tv;

  double* sum1(dataStart);
  double* sum2(v.dataStart);
  double* dest(tv.dataStart);

  *(dest++)=*(sum1++) + *(sum2++);
  *(dest++)=*(sum1++) + *(sum2++);
  *dest=*sum1 + *sum2;
  
  return tv;
}


Inline TVector<double, 3>
Vector<double, 3>::operator-(const Vector<double, 3>& v) const
{
  TVector<double, 3> tv;

  double* sum1(dataStart);
  double* sum2(v.dataStart);
  double* dest(tv.dataStart);

  *(dest++)=*(sum1++) - *(sum2++);
  *(dest++)=*(sum1++) - *(sum2++);
  *dest=*sum1 - *sum2;
  
  return tv;
}

Inline TVector<double, 3>
Vector<double, 3>::operator*(double t) const
{
  TVector<double, 3> tv(*this);
  double* ptr(tv.dataStart);

  *(ptr++) *= t;
  *(ptr++) *= t;
  *(ptr) *= t;
  
  
  return tv;
}


Inline double
Vector<double, 3>::operator*(const Vector<double, 3> &v) const
{
  double result;

  double* count1(dataStart);
  double* count2(v.dataStart);

  result =*(count1++) * *(count2++);
  result+=*(count1++) * *(count2++);
  result+=*count1 * *count2;
  
  return result;
}


Inline double
Vector<double, 3>::operator*(const TVector<double, 3> &v) const
{
  double result;

  double* count1(dataStart);
  double* count2(v.dataStart);

  result =*(count1++) * *(count2++);
  result+=*(count1++) * *(count2++);
  result+=*count1 * *count2;

  allocator_data.deallocate(v.dataStart);
  return result;
}


Inline double
TVector<double, 3>::operator*(const Vector<double, 3> &v) const
{
  double result;

  double* count1(dataStart);
  double* count2(v.dataStart);

  result =*(count1++) * *(count2++);
  result+=*(count1++) * *(count2++);
  result+=*count1 * *count2;
  
  allocator_data.deallocate(dataStart);
  return result;
}


Inline double
TVector<double, 3>::operator*(const TVector<double, 3> &v) const
{
  double result;

  double* count1(dataStart);
  double* count2(v.dataStart);

  result =*(count1++) * *(count2++);
  result+=*(count1++) * *(count2++);
  result+=*count1 * *count2;
  
  allocator_data.deallocate(dataStart);
  allocator_data.deallocate(v.dataStart);
  return result;
}

#else


  
template <class T, int N>
Inline TVector<T,N> TVector<T, N>::operator+(const TVector<T, N>& tv) const
{
  T* sum1(dataStart);
  T* sum2(tv.dataStart);

  do
    *(sum1++) += *(sum2++); 
  while (sum1 != dataFinish);
    

  allocator_data.deallocate(tv.dataStart);
  
  return *this;
}



template <class T, int N>
Inline TVector<T,N> TVector<T, N>::operator+(const Vector<T, N>& v) const
{
  
  T* sum1(dataStart);
  T* sum2(v.dataStart);

  do
    *(sum1++) += *(sum2++);    
  while (sum1 != dataFinish);

  return *this;
}

template <class T, int N>
Inline TVector<T,N> TVector<T, N>::operator*(T t) const
{
  
  T* ptr(dataStart);
  do
    *(ptr++) *= t;
  while (ptr != dataFinish);

  return *this;
}

template <class T, int N>
Inline TVector<T, N> Vector<T, N>::operator*(T t) const
{
  TVector<T, N> tv(*this);

  T* ptr(tv.dataStart);
  do
    *(ptr++) *= t;
  while (ptr != tv.dataFinish);
  
  return tv;
}





template <class T, int N>
Inline TVector<T,N> TVector<T, N>::operator-(const TVector<T, N>& tv) const
{
  T* sum1(dataStart);
  T* sum2(tv.dataStart);

  do
    *(sum1++) -= *(sum2++); 
  while (sum1 != dataFinish);
    

  allocator_data.deallocate(tv.dataStart);
  
  return *this;
}



template <class T, int N>
Inline TVector<T,N> TVector<T, N>::operator-(const Vector<T, N>& v) const
{
  
  T* sum1(dataStart);
  T* sum2(v.dataStart);

  do
    *(sum1++) -= *(sum2++);    
  while (sum1 != dataFinish);

  return *this;
}


template <class T, int N>
Inline Vector<T, N>::Vector(const Vector<T, N>& v)
           : TVector<T, N>(allocator_data.allocate())
{
  T* dest(dataStart);
  T* src(v.dataStart);

  do
    *(dest++)=*(src++);
  while (dest != dataFinish);
}



template <class T, int N>
Inline Vector<T, N>& Vector<T, N>::operator=(const Vector<T, N>& v)
{
  T* dest(dataStart);
  T* src(v.dataStart);

  while (dest != dataFinish)
    *(dest++)=*(src++);

  return *this;
}


template <class T, int N>
Inline Vector<T, N>& Vector<T, N>::operator=(const TVector<T, N>& tv)
{
  allocator_data.deallocate(dataStart);

  dataStart=tv.dataStart;
  dataFinish  =tv.dataFinish;

  return *this;
}



template <class T, int N>
Inline TVector<T, N> Vector<T, N>::operator+(const Vector<T, N>& v) const
{
  TVector<T, N> tv;

  T* sum1(dataStart);
  T* sum2(v.dataStart);
  T* dest(tv.dataStart);

  do
    *(dest++)=*(sum1++) + *(sum2++);
  while (sum1 != dataFinish);
  
  return tv;
}




template <class T, int N>
Inline TVector<T, N> Vector<T, N>::operator-(const Vector<T, N>& v) const
{
  TVector<T, N> tv;

  T* sum1(dataStart);
  T* sum2(v.dataStart);
  T* dest(tv.dataStart);

  do
    *(dest++)=*(sum1++) - *(sum2++);
  while (sum1 != dataFinish);
  
  return tv;
}



template <class T, int N>
Inline T Vector<T, N>::operator*(const Vector<T, N> &v) const
{
  T result(0);

  T* count1(dataStart);
  T* count2(v.dataStart);

  while (count1 != dataFinish)
    result+=*(count1++) * *(count2++);

  return result;
}


template <class T, int N>
Inline T Vector<T, N>::operator*(const TVector<T, N> &tv) const
{
  T result(0);

  T* count1(dataStart);
  T* count2(tv.dataStart);

  while (count1 != dataFinish)
    result+=*(count1++) * *(count2++);

  allocator_data.deallocate(tv.dataStart);
  
  return result;
}


template <class T, int N>
Inline T TVector<T, N>::operator*(const Vector<T, N> &v) const
{
  T result(0);

  T* count1(dataStart);
  T* count2(v.dataStart);

  while (count1 != dataFinish)
    result+=*(count1++) * *(count2++);

  allocator_data.deallocate(dataStart);
  
  return result;
}



template <class T, int N>
Inline T TVector<T, N>::operator*(const TVector<T, N> &tv) const
{
  T result(0);

  T* count1(dataStart);
  T* count2(tv.dataStart);

  while (count1 != dataFinish)
    result+=*(count1++) * *(count2++);

  allocator_data.deallocate(tv.dataStart);
  allocator_data.deallocate(dataStart);

  return result;
}
#endif



template <class T, int N>
Inline TVector<T,N>::TVector(T* start) 
{ 
  dataStart=start;
  dataFinish=start+N; 
};
 
 
#endif

