
/* $Id: Alloc.h,v 0.2.1.1 1996/08/01 11:25:00 spander Exp spander $ */

/*
 * $Log: Alloc.h,v $
 * Revision 0.2.1.1  1996/08/01 11:25:00  spander
 * starting branch 0.2.1
 *
 * Revision 0.2  1996/08/01 11:25:00  spander
 * pre ALPHA
 *
 */

/*                                                                    */ 
/* Allocator class                                                    */
/*                                                                    */
/* York Xylander & Klaus Spanderen, 28.06.96                          */

#ifndef ALLOC_H
#define ALLOC_H
#define FALSE 0

#include <stdlib.h>
#include <iostream.h>
#include <stdio.h>

template <class T, int N>
class StaticStore
{
private:
  union Magic {
    T*  content;
    T** nextFree;
  } lastFree;

public:  
  StaticStore(void);
  StaticStore(const int n);
  StaticStore(const int n, T* const mem);
  ~StaticStore();

         T*   addMem    (const int n, T* const mem);
  inline T*   allocate  (void);
  inline void deallocate(T* const s);
};


/* Template Metaprogramming to detect storage errors at compile time  */

template <int condition>
class StoreError 
{
public:
  static inline void f() {};
};

class StoreError<0>
{
public:
  static inline void f()
  {
    cout << "Can't create Storage manager" << endl;
    cout << "Program abort" << endl; 
    exit(-1);
  }
};

/* end Template Metaprogamming                                        */


template <class T, int N>
inline StaticStore<T, N>::StaticStore(void)
{
  lastFree.content = NULL;

#ifndef CRAY_T3D
  enum { condition=(N*sizeof(T) >= sizeof(T*)) };
  StoreError<condition>::f();
#endif
}


template <class T, int N>
StaticStore<T, N>::StaticStore(const int n)
{
#ifndef CRAY_T3D
  enum { condition=(N*sizeof(T) >= sizeof(T*)) };
  StoreError<condition>::f();
#endif  

  lastFree.content=(T*) malloc(n*N*sizeof(T));
  
  T** tmp;
  T*  run(lastFree.content);
  T*  end(lastFree.content+(n-1)*N);

  while (run < end)
    {
      tmp  = (T**) run;
      run += N;
      *tmp = run;
    }

  *((T**) run) = NULL;
}


template <class T, int N>
StaticStore<T, N>::StaticStore(const int n, T* const mem)
{
#ifndef CRAY_T3D
  enum { condition=(N*sizeof(T) >= sizeof(T*)) };
  StoreError<condition>::f();
#endif
  lastFree.content=mem;

  T** tmp;
  T*  run(lastFree.content);
  T*  end(lastFree.content+(n-1)*N);

  while (run < end)
    {
      tmp  = (T**) run;
      run +=N;
      *tmp = run;
    }

  *((T**) run) = NULL;
}


template <class T, int N>
T* StaticStore<T, N>::addMem(const int n, T* const mem)
{
  T* oldEnd(lastFree.content);

  lastFree.content=mem;

  T*  run(lastFree.content);
  T*  end(lastFree.content+(n-1)*N);

  while (run < end)
  *((T**) run) = (run+=N);

  *((T**) run) = oldEnd;

  return mem;
}  
  

template <class T, int N>
inline StaticStore<T, N>::~StaticStore()
{
  /*
     first point of an imaginary todo list: write a destructor
  */
}


template <class T, int N>
inline void StaticStore<T, N>::deallocate(T* const s)
{
  T* const tmp(s);
  *((T**) s)=lastFree.content;
  lastFree.content=tmp;
}


template <class T, int N>
inline T* StaticStore<T, N>::allocate(void)
{
  if (lastFree.content != NULL)
    {
      T* s             = lastFree.content;
      lastFree.content =*lastFree.nextFree;
      return s;
    }
  else
  return (T*) malloc(N*sizeof(T));
}


#endif 
// Local Variables:
// mode: c++
// End:





