
/* $Id: Vector.h,v 0.2.1.1 1996/08/01 11:25:01 spander Exp spander $ */

/*
 * $Log: Vector.h,v $
 * Revision 0.2.1.1  1996/08/01 11:25:01  spander
 * starting branch 0.2.1
 *
 * Revision 0.2  1996/08/01 11:25:01  spander
 * pre ALPHA
 *
 */


 
/*                                                                  */ 
/* Vector class                                                     */
/* Lit: Advanced C++, James O. Coplien                              */
/*      C++ Report (May 1995) "Using C++ template metaprograms"     */
/*                                                                  */





#ifndef VECTOR_H
#define VECTOR_H

#include <stdarg.h>
#include "Alloc.h"



static const void   *src1;
static const void   *src2;
static       void   *dest;


template<class T, int N>
class VectorMem
{
protected:
  T* dataStart;
  VectorMem(T* const _dataStart) : dataStart(_dataStart) { }

  static StaticStore<T, N> vectorStore;

public:
  static T* addMem(const int n, T* const mem) 
    { return vectorStore.addMem(n, mem); }
};


template <class T, int N> class Vector;

template <class T, int N>
class TmpVector : public VectorMem<T, N> 
{                                        
private:
  TmpVector() : VectorMem<T, N>(vectorStore.allocate()) { }
  TmpVector(const TmpVector<T, N>& tv) 
    : VectorMem<T, N>(tv.dataStart) { }

public:
         T&            operator[](int n)       { return *(dataStart+n); }
         T             operator()(int n) const { return *(dataStart+n); }
         T*            mem(void)               { return dataStart; }

  inline TmpVector<T, N> operator+(const Vector<T, N>& v) const;
  inline TmpVector<T, N> operator+(const TmpVector<T, N>& tv) const;
  
  inline TmpVector<T, N> operator-(const Vector<T, N>& v) const;
  inline TmpVector<T, N> operator-(const TmpVector<T, N>& tv) const;

  inline TmpVector<T, N> operator*(const T t);
  inline T               operator*(const Vector<T, N>& v) const;
  inline T               operator*(const TmpVector<T, N>& tv) const;

  friend class Vector<T, N>;
};



template <class T, int N>
class Vector : public VectorMem<T, N>  
{                                      
public:
  typedef T type;

  Vector() : VectorMem<T, N>(vectorStore.allocate()) { }
  Vector(T t, ... );
  Vector(const Vector<T, N>& v);

  ~Vector() { vectorStore.deallocate(dataStart); }
  
         T&            operator[](int n)       { return *(dataStart+n); }
         T             operator()(int n) const { return *(dataStart+n); }
         T*            mem(void)               { return dataStart; }

  inline Vector<T, N>&   operator=(const Vector<T, N>& v);
  inline Vector<T, N>&   operator=(const TmpVector<T, N>& tv);
  inline Vector<T, N>&   operator=(const T& t);

  inline Vector<T, N>    operator+=(const Vector<T, N>& v);
  inline Vector<T, N>    operator+=(const TmpVector<T, N>& tv);

  inline TmpVector<T, N> operator+(const Vector<T, N>& v) const;
  inline TmpVector<T, N> operator+(const TmpVector<T, N>& tv) const;

  inline TmpVector<T, N> operator-(const Vector<T, N>& v) const;
  inline TmpVector<T, N> operator-(const TmpVector<T, N>& tv) const;

  inline TmpVector<T, N> operator*(const T t);
  inline T               operator*(const Vector<T, N>& v) const;
  inline T               operator*(const TmpVector<T, N>& tv) const;

  friend class TmpVector<T, N>;
};




/* start Template Metaprogramming                                         */

template <class T, int I>
class VecEqual
{
public:
  static inline void loop()
  {
    *((T*) dest)=*((T*) src1);
    dest=(void*) ((char*) dest+sizeof(T));
    src1=(void*) ((char*) src1+sizeof(T));
    VecEqual<T, (I-1)>::loop();
  }
};

class VecEqual<double, 1>
{
public:
  static inline void loop() { *((double*) dest)=*(double*)(src1); }
};


template <class T, int I>
class VecScaMultiInPlace
{
public:
  static inline void loop(const T t)
  {
    *((T*) dest)*=t;
    dest=(void*) ((char*) dest+sizeof(T));
    VecScaMultiInPlace<T, (I-1)>::loop(t);
  }
};

class VecScaMultiInPlace<double, 1>
{
public:
  static inline void loop(const double t) { *((double*) dest)*=t; }
};


template <class T, int I>
class VecAdd
{
public:
  static inline void loop()
  {
    *((T*) dest)=*((T*) src1) + *((T*) src2);
    dest=(void*) ((char*) dest+sizeof(T));
    src1=(void*) ((char*) src1+sizeof(T));
    src2=(void*) ((char*) src2+sizeof(T));
    VecAdd<T, (I-1)>::loop();
  }
};

class VecAdd<double, 1>
{
public:
  static inline void loop() 
  { *((double*) dest)=*(double*)(src1) + *((double*) src2); }
};


template <class T, int I>
class VecSub
{
public:
  static inline void loop()
  {
    *((T*) dest)=*((T*) src1) - *((T*) src2);
    dest=(void*) ((char*) dest+sizeof(T));
    src1=(void*) ((char*) src1+sizeof(T));
    src2=(void*) ((char*) src2+sizeof(T));
    VecSub<T, (I-1)>::loop();
  }
};

class VecSub<double, 1>
{
public:
  static inline void loop()
  { *((double*) dest)=*(double*)(src1) - *((double*) src2); }
};


template <class T, int I>
class VecAddinPlace
{
public:
  static inline void loop()
  {
    *((T*) dest) += *((T*) src1);
    dest=(void*) ((char*) dest+sizeof(T));
    src1=(void*) ((char*) src1+sizeof(T));
    VecAddinPlace<T, (I-1)>::loop();
  }
};

class VecAddinPlace<double, 1>
{
public:
  static inline void loop() { *((double*) dest)+=*(double*)(src1); }
};


template <class T, int I>
class VecSubinPlace
{
public:
  static inline void loop()
  {
    *((T*) dest)-=*((T*) src1);
    dest=(void*) ((char*) dest+sizeof(T));
    src1=(void*) ((char*) src1+sizeof(T));
    VecSubinPlace<T, (I-1)>::loop();
  }
};

class VecSubinPlace<double, 1>
{
public:
  static inline void loop() { *((double*) dest) -=*(double*)(src1); }
};


template <class T, int I>
class VecScaMulti
{
public:
  static inline void loop(const T t)
  {
    *((T*) dest)=*((T*) src1) * t;
    dest=(void*) ((char*) dest+sizeof(T));
    src1=(void*) ((char*) src1+sizeof(T));
    VecScaMulti<T, (I-1)>::loop(t);
  }
};

class VecScaMulti<double, 1>
{
public:
  static inline void loop(const double t)
    { *((double*) dest)=*((double*) src1) * t;}
};



template <class T, int I>
class VecScaEqual
{
public:
  static inline void loop(const T& t)
  {
    *((T*) dest)=t;
    dest=(void*) ((char*) dest+sizeof(T));    
    VecScaMulti<T, (I-1)>::loop(t);
  }
};

class VecScaEqual<double, 1>
{
public:
  static inline void loop(const double& t) { *((double*) dest)=t;}
};



template <class T, int I>
class VecBiMulti
{
public:
  static inline void loop(T& t)
  {
    t  += *((T*) src1) * *((T*) src2);
    
    src1= (void*) ((char*) src1+sizeof(T));
    src2= (void*) ((char*) src2+sizeof(T));
    
    VecBiMulti<T, (I-1)>::loop(t);
  }
};


class VecBiMulti<double, 1>
{
public:
  static inline void loop(double& t)
  { t  += *((double*) src1) * *((double*) src2); }
};


/* end Template Metaprogramming                                           */


template <class T, int N>
Vector<T, N>::Vector(T t, ... )
  : VectorMem<T, N>(vectorStore.allocate())
{
  dataStart[0]=t;
  va_list arguments;
  
  va_start(arguments, t);
  for (int i=1; i<N; ++i) 
    dataStart[i]=va_arg(arguments, T);
  va_end(arguments);
}


template <class T, int N>
Vector<T, N>::Vector(const Vector<T, N>& v)
  : VectorMem<T, N>(vectorStore.allocate())
{
  dest=dataStart;
  src1=v.dataStart;

  VecEqual<T, N>::loop();
}


template <class T, int N>
inline Vector<T, N>& Vector<T, N>::operator=(const Vector<T, N>& v)
{
  src1=v.dataStart;
  dest=dataStart;

  VecEqual<T, N>::loop();
  return *this;
}


template <class T, int N>
inline Vector<T, N>& Vector<T, N>::operator=(const TmpVector<T, N>& tv)
{
#ifndef EQUAL_IN_PLACE
  vectorStore.deallocate(dataStart);
  dataStart=tv.dataStart;
#else
  src1=tv.dataStart;
  dest=dataStart;

  VecEqual<T, N>::loop();
  vectorStore.deallocate(tv.dataStart);
#endif

  return *this;
}


template <class T, int N>
inline Vector<T, N>& Vector<T, N>::operator=(const T& t)
{
  dest=dataStart;
  VecScaEqual<T, N>::loop(t);

  return *this;
}


template <class T, int N> 
inline TmpVector<T, N> Vector<T, N>::operator+(const Vector<T, N>& v) const
{
  TmpVector<T, N> tv; 
  dest=tv.dataStart;
  src1=v.dataStart;
  src2=dataStart;
  
  VecAdd<T, N>::loop();
  return tv;
}


template <class T, int N> inline TmpVector<T, N> 
Vector<T, N>::operator+(const TmpVector<T, N>& tv) const
{ 
  return tv + *this; 
}


template <class T, int N>
inline TmpVector<T, N> Vector<T, N>::operator-(const Vector<T, N> &v) const
{
  TmpVector<T, N> tv;
  dest=tv.dataStart;
  src1=dataStart;
  src2=v.dataStart;

  VecSub<T, N>::loop();
  return tv;
}


template <class T, int N>
inline TmpVector<T, N> 
Vector<T, N>::operator-(const TmpVector<T, N> &tv) const
{
  dest=tv.dataStart;
  src1=dataStart;

  VecSubinPlace<T, N>::loop();
  return tv;
}


template <class T, int N> inline TmpVector<T, N> 
TmpVector<T, N>::operator+(const Vector<T, N>& v) const
{
  src1=v.dataStart;
  dest=dataStart;

  VecAddinPlace<T, N>::loop();
  return *this;
}


template <class T, int N>
inline Vector<T, N> Vector<T, N>::operator+=(const Vector<T, N>& v)
{
  src1 = v.dataStart;
  dest = dataStart;
  VecAddinPlace<T, N>::loop();
  
  return *this;
}


template <class T, int N>
inline Vector<T, N> Vector<T, N>::operator+=(const TmpVector<T, N>& tv)
{
  src1=tv.dataStart;
  dest=dataStart;
  VecAddinPlace<T, N>::loop();
  
  vectorStore.deallocate(tv.dataStart);
  return *this;
}  


template <class T, int N> inline TmpVector<T, N> 
TmpVector<T, N>::operator+(const TmpVector<T, N>& tv) const
{
  src1=tv.dataStart;
  dest=dataStart;

  VecAddinPlace<T, N>::loop();
  vectorStore.deallocate(tv.dataStart);
  return *this;
}


template <class T, int N> inline TmpVector<T, N> 
TmpVector<T, N>::operator-(const Vector<T, N>& v) const
{
  src1=v.dataStart;
  dest=dataStart;

  VecSubinPlace<T, N>::loop();
  return *this;
}


template <class T, int N> inline TmpVector<T, N> 
TmpVector<T, N>::operator-(const TmpVector<T, N>& tv) const
{
  src1=tv.dataStart;
  dest=dataStart;

  VecSubinPlace<T, N>::loop();
  vectorStore.deallocate(tv.dataStart);
  return *this;
}


template <class T, int N>
inline TmpVector<T, N> Vector<T, N>::operator*(const T t)
{
  TmpVector<T, N> tv;
  dest=tv.dataStart;
  src1=dataStart;
  
  VecScaMulti<T, N>::loop(t);
  return tv;
}


template <class T, int N>
inline TmpVector<T, N> TmpVector<T, N>::operator*(const T t)
{
  dest=dataStart;

  VecScaMultiInPlace<T, N>::loop(t);
  return *this;
}


template <class T, int N>
inline T TmpVector<T, N>::operator*(const Vector<T, N>& v) const
{
  src1  =dataStart;
  src2  =v.dataStart;
  
  T result(0);
  VecBiMulti<T, N>::loop(result);
  
  vectorStore.deallocate(dataStart);
  return result;
}


template <class T, int N>
inline T Vector<T, N>::operator*(const TmpVector<T, N>& tv) const
{
  return tv * *this; 
}


template <class T, int N>
inline T TmpVector<T, N>::operator*(const TmpVector<T, N>& tv) const
{
  src1  =dataStart;
  src2  =tv.dataStart;

  T result(0);
  VecBiMulti<T, N>::loop(result);
  
  vectorStore.deallocate(dataStart);
  vectorStore.deallocate(tv.dataStart);
  return result;
}


template <class T, int N>
inline T Vector<T, N>::operator*(const Vector<T, N> &v) const
{
  src1  =dataStart;
  src2  =v.dataStart;

  T result;
  VecBiMulti<T, N>::loop(result);

  return result;
}
   


#endif
// Local Variables:
// mode: c++
// End:

