#include <stdlib.h>
#include <math.h>
#include <stdio.h>


int T, L, N;

double ZERO=0.0;
double ONE=1.0;

#ifdef NAG
double g05daf(double* a, double* b);
#define drand48() g05daf(&ZERO,&ONE)
#endif

double *sigma0,*sigma1,*sigma2;
double sigma0Sum,sigma1Sum,sigma2Sum;
double beta,deltaS;
double sigma0Trial,sigma1Trial,sigma2Trial;

int *neighborN, *neighborW, *neighborS, *neighborE;


int Accept, reject;

void initSigma(int timeExtension, int spaceExtension, double b);
void exitSigma();

void randomize3(double* s0, double* s1, double* s2);
int index(int t, int x); 
void sumNeighbors(int i);
void metropolisUpdate(int sweeps, int hits);
void localMetropolisUpdate(int i, int hits);
double measureLinkEnergy();


#define index(t,x) L*t+x
/*
int index(int t, int x)
{
    return L*t+x;
}
*/

void randomize3(double* s0,double* s1,double* s2)
{
    double r;
    do { 
        *s0=2.0*drand48()-1.0; 
        *s1=2.0*drand48()-1.0; 
        *s2=2.0*drand48()-1.0; 
        r=(*s0)*(*s0)+(*s1)*(*s1)+(*s2)*(*s2); 
    } while (r >= 1.0 || r == 0.0);
    r=sqrt(r);
    *s0 /= r;
    *s1 /= r;
    *s2 /= r;
}

void initSigma(int timeExtension,
	       int spaceExtension,
	       double b)
{
    int t, x, i;
    int tup, tdown, xup, xdown;

    L=spaceExtension;
    T=timeExtension;
    N=(L*T);
    beta=b;
    Accept=0;
    reject=0;

    sigma0 = (double*) malloc(sizeof(double)*N);
    sigma1 = (double*) malloc(sizeof(double)*N);
    sigma2 = (double*) malloc(sizeof(double)*N);

    neighborE = (int*) malloc(sizeof(int)*N);
    neighborN = (int*) malloc(sizeof(int)*N);
    neighborW = (int*) malloc(sizeof(int)*N);
    neighborS = (int*) malloc(sizeof(int)*N);

    for (t=0;t<T;t++)
	for (x=0;x<L;x++)
	    {
		i = index(t,x);
		randomize3(&sigma0[i],
			   &sigma1[i],
			   &sigma2[i]);
			   
	        tup = (t+1)%T;
		tdown = (t-1+T)%T;
		xup = (x+1)%L;
		xdown = (x-1+L)%L;


		
		neighborE[i]=index(tup,x);
		neighborN[i]=index(t,xup);
		neighborW[i]=index(tdown,x);
		neighborS[i]=index(t,xdown);
	    }
}


void exitSigma()
{
     free(sigma0);
     free(sigma1);
     free(sigma2);
     free(neighborE);
     free(neighborN);
     free(neighborW);
     free(neighborS);
}

void sumNeighbors(int i)
{
    int iE,iN,iW,iS;

    iE = neighborE[i];		
    iN = neighborN[i];		
    iW = neighborW[i];		
    iS = neighborS[i];		
    
    sigma0Sum=sigma0[iE]+sigma0[iN]+sigma0[iW]+sigma0[iS];
    sigma1Sum=sigma1[iE]+sigma1[iN]+sigma1[iW]+sigma1[iS];
    sigma2Sum=sigma2[iE]+sigma2[iN]+sigma2[iW]+sigma2[iS];
}

void localMetropolisUpdate(int i, int hits)
{
    sumNeighbors(i);

    while (hits--)
	{
	    randomize3(&sigma0Trial,
		       &sigma1Trial,
		       &sigma2Trial);
	    deltaS = 
		-beta * ((sigma0Trial-sigma0[i])*sigma0Sum+
			 (sigma1Trial-sigma1[i])*sigma1Sum+
			 (sigma2Trial-sigma2[i])*sigma2Sum);

	    if (deltaS<0 || 
		drand48() < exp(-deltaS))
		{
		    sigma0[i]=sigma0Trial;
		    sigma1[i]=sigma1Trial;
		    sigma2[i]=sigma2Trial;
		    Accept++;
		}
	    else
		{
		    reject++;
		}

	}
}


void metropolisUpdate(int sweeps, int hits)
{
    int i;
    while (sweeps--)
	{
	    for (i=0;i<N;i++)
		localMetropolisUpdate(i,hits);
	}
}

double measureLinkEnergy()
{
    double link=0;
    int t,x,i,iE,iN,iW,iS;

    for (t=0;t<T;t++)
	for (x=0;x<L;x++)
	    {
		i = index(t,x);
 
		
		iE = neighborE[i];		
		iN = neighborN[i];		
		iW = neighborW[i];		
		iS = neighborS[i];		


		link+=
		    sigma0[i]*(sigma0[iE]+sigma0[iN]+sigma0[iW]+sigma0[iS])+
		    sigma1[i]*(sigma1[iE]+sigma1[iN]+sigma1[iW]+sigma1[iS])+
		    sigma2[i]*(sigma2[iE]+sigma2[iN]+sigma2[iW]+sigma2[iS]);
	    }
    link /= N;
    return link;
}


main()
{
    int equi=1000;
    int sweeps=1000,s;
    double meanLink=0;

    initSigma(10,10,1.0);
    metropolisUpdate(equi,4);

    for (s=0;s<sweeps;s++)
	{
	    metropolisUpdate(1,4);
	    meanLink+=measureLinkEnergy();
	}
    meanLink /= sweeps;
    printf("mean=%G\n",meanLink);
    printf("ratio=%G\n",Accept/(double)(Accept+reject));
}



